-- RDC.e
-- Remote Dynamic Calls
-- Pete Eberlein <xseal@harborside.com>
-- Oct. 2 1999
-- version 0.1 -- first release

--- minor mods made by Irv Mullins 02/14/00 
--- to comment out some print lines

-- Description:
--   Allows one Euphoria program to communicate with another Euphoria program,
-- not necessarily running on the same computer, using routines similar to the 
-- standard Euphoria dynamic call interfaces: routine_id, call_proc, call_func.
-- The routines are based on a client-server model, the server exports its 
-- functions and procedures, which the client can connect to and call its 
-- routines.  The server can currently only accomodate one client at a time.

-- Server procedures:
--  export_routine
--  serve_routines

-- Client routines:
--  sd = remote_open("hostname.domain")
--  id = remote_routine_id(sd, "routine_name")
--  result = remote_call_func(sd, id, {arguments})
--  remote_call_proc(sd, id, {arguments})
--  remote_close(sd)

without warning
include dll.e
include machine.e

global integer default_port
default_port = 10578

constant
AF_INET = 2,
INADDR_ANY = 0,
SOCK_STREAM = 1,
DEFAULT_PROTOCOL = 0,
sizeof_sockaddr = 16,
SIGINT = 2,
SIGTERM = 15


constant lib = open_dll("")

constant 
accept = define_c_func(lib, "accept", {C_INT, C_POINTER, C_POINTER}, C_INT),
bind = define_c_func(lib, "bind", {C_INT, C_POINTER, C_POINTER}, C_INT),
close_ = define_c_proc(lib, "close", {C_INT}),
connect = define_c_func(lib, "connect", {C_INT, C_POINTER, C_INT}, C_INT),
fork = define_c_func(lib, "fork", {}, C_INT),
listen = define_c_func(lib, "listen", {C_INT, C_INT}, C_INT),
read = define_c_func(lib, "read", {C_INT, C_POINTER, C_INT}, C_INT),
signal = define_c_func(lib, "signal", {C_INT, C_POINTER}, C_INT),
socket = define_c_func(lib, "socket", {C_INT, C_INT, C_INT}, C_INT),
wait = define_c_func(lib, "wait", {C_POINTER}, C_INT),
waitpid = define_c_func(lib, "waitpid", {C_INT, C_POINTER, C_INT}, C_INT),
write = define_c_func(lib, "write", {C_INT, C_POINTER, C_INT}, C_INT),

gethostbyname = define_c_func(lib, "gethostbyname", {C_POINTER}, C_POINTER),
inet_ntoa = define_c_func(lib, "inet_ntoa", {C_POINTER}, C_POINTER),
inet_aton = define_c_func(lib, "inet_aton", {C_POINTER,C_POINTER}, C_INT),
inet_addr = define_c_func(lib, "inet_ntoa", {C_POINTER}, C_POINTER),
htonl = define_c_func(lib, "htonl", {C_ULONG}, C_ULONG),
htons = define_c_func(lib, "htons", {C_USHORT}, C_USHORT),
ntohl = define_c_func(lib, "ntohl", {C_ULONG}, C_ULONG),
ntohs = define_c_func(lib, "ntohs", {C_USHORT}, C_USHORT),

perror = define_c_proc(lib, "perror", {C_POINTER}),
sleep = define_c_func(lib, "sleep", {C_INT}, C_INT)

if find(-1, {
accept,
bind,
connect,
fork,
read,
socket,
wait,
waitpid,
write,
gethostbyname,
htonl,
htons,
ntohl,
ntohs,
perror,
sleep
}) then
   puts(2, "required shared routine missing\n")
end if

-------------------------------------------------------------------------

sequence strings strings = {}

function alloc_string(sequence str)
    atom a
    a = allocate(length(str)+1)
    poke(a, str & 0)
    strings = append(strings, a)
    return a
end function

procedure free_strings(integer count)
    for i = length(strings)-count+1 to length(strings) do
    	free(strings[i])
    end for
    strings = strings[1..length(strings)-count]
end procedure

------------------------------------------------------------------------

function cleanExit()
    -- automatically closes any open file or socket descriptors
    puts(1, "cleanExit()\n")
    abort(0)
    return NULL
end function

--------------------------------------------------------------------------

constant data = allocate(12)
integer error

function read_object(integer fd)
    object result
    integer len
    if 4 != c_func(read, {fd, data, 4}) then
    	error = 1
    	return -1
    end if

    if peek(data+3) = #80 then
    	-- sequence
    	len = and_bits(peek4u(data), #FFFFFF)
        result = repeat(0, len)
    	for i = 1 to len do
    	    result[i] = read_object(fd)
    	end for
        return result
    end if
    if peek(data+3) = #40 then
    	-- atom
        if 8 != c_func(read, {fd, data, 8}) then
            error = 1
    	    return -1
        end if
	return float64_to_atom(peek({data, 8}))
    end if

    return peek4s(data)
end function

---------------------------------------------------------------------------

procedure write_object(integer fd, object o)
    if integer(o) then
    	poke4(data, o)
	if 4 != c_func(write, {fd, data, 4}) then
	end if
    elsif atom(o) then	
    	poke(data+4, atom_to_float64(o))
    	poke(data+3, #40)
	if 12 != c_func(write, {fd, data, 12}) then
	end if
    else --if sequence(o) then	
    	poke4(data, length(o))
    	poke(data+3, #80)
	if 4 != c_func(write, {fd, data, 4}) then
	end if
	for i = 1 to length(o) do
	    write_object(fd, o[i])
	end for
    end if
end procedure

---------------------------------------------------------------------------

sequence routine_names, routine_ids
routine_names = {}
routine_ids = {}

global procedure export_routine(sequence name, integer id)
    routine_names &= {name}
    routine_ids &= id
end procedure

---------------------------------------------------------------------------
atom serverSocket, clientSocket

global procedure serve_routines()
    integer sd, cd, id
    object command

    -- create a socket
    sd = c_func(socket, {AF_INET, SOCK_STREAM, DEFAULT_PROTOCOL})
    if sd = -1 then
    	c_proc(perror, {alloc_string("socket")})
	free_strings(1)
	abort(1)
    end if
---    printf(1, "server descriptor = %d\n", {sd})
    
    -- create the server socket structure
    serverSocket = allocate(sizeof_sockaddr)
    poke4(serverSocket, {
    	AF_INET +
	c_func(htons, {default_port}) * #10000,
	c_func(htonl, {INADDR_ANY}),
	0,
	0})

    -- bind the socket to its port
    if -1 = c_func(bind, {sd, serverSocket, sizeof_sockaddr}) then
    	c_proc(perror, {alloc_string("bind")})
	free_strings(1)
	abort(1)
    end if
    
    -- start listening for connections, queue up to 20
    if -1 = c_func(listen, {sd, 20}) then
    	c_proc(perror, {alloc_string("listen")})
	free_strings(1)
	abort(1)
    end if
    
    -- set up ctrl-c handlers
    if c_func(signal, {SIGINT, call_back(routine_id("cleanExit"))}) then
    end if
    if c_func(signal, {SIGTERM, call_back(routine_id("cleanExit"))}) then
    end if

    -- create client socket structure
    clientSocket = allocate(sizeof_sockaddr)
    while 1 do

---	printf(1, "waiting for a client...\n", {})

	-- wait for a client to connect
	cd = c_func(accept, {sd, clientSocket, sizeof_sockaddr})
	if cd = -1 then
    	    c_proc(perror, {alloc_string("accept")})
	    free_strings(1)
	    abort(1)
	end if


---	printf(1, "connect %d\n", cd)
	while 1 do

	    -- read a command object
	    error = 0
    	    command = read_object(cd)
	    if error then
		exit
	    end if

    	    if sequence(command) then
		-- request for a routine id
		id = find(command, routine_names)
		if id then
    	    	    write_object(cd, routine_ids[id])
   ---              printf(1,"Exporting %s as #%d\n",{command,routine_ids[id]})
		else
    	    	    write_object(cd, -1)
   ---              printf(1,"Returned id for routine %s\n",{command})
		end if

	    elsif command > 0 then
		-- call a procedure
		call_proc(command-1, read_object(cd))

	    else
		-- call a function
		write_object(cd, call_func(-command, read_object(cd)))
 
	    end if
	end while
---	printf(1, "close %d\n", {cd})
	c_proc(close_, {cd})

    end while

    free(clientSocket)
    free(serverSocket)
    
    c_proc(close_, {sd})
    
end procedure

----------------------------------------------------------------------------

global function remote_open(sequence hostname)
    -- returns a socket descriptor, or -1 on error

    integer sd
    atom addr, hoststruct, hostnode
    
    if length(hostname) = 0 then
    	-- empty string: use any host
    	addr = c_func(htonl, {INADDR_ANY})

    elsif find(hostname[1], "0123456789") then
    	-- dotted quad
        hostnode = allocate(4)
    	if not c_func(inet_aton, {alloc_string(hostname), hostnode}) then
    	    free_strings(1)
    	    free(hostnode)
	    return -1
	end if 
	addr = peek4u(hostnode)
	free_strings(1)
	free(hostnode)

    else
    	-- name lookup
    	hoststruct = c_func(gethostbyname, {alloc_string(hostname)})
	free_strings(1)
	if not hoststruct then
	    return -1  -- host not found
	end if
	hostnode = peek4u(peek4u(hoststruct + 16))
	addr = peek4u(hostnode)

    end if

    -- create the socket structure
    serverSocket = allocate(sizeof_sockaddr)
    poke4(serverSocket, {
    	AF_INET +
	c_func(htons, {default_port}) * #10000,
	addr,
	0,
	0})
	
    -- create the socket descriptor
    sd = c_func(socket, {AF_INET, SOCK_STREAM, DEFAULT_PROTOCOL})
    if sd = -1 then
    	c_proc(perror, {alloc_string("socket")})
	free_strings(1)
	free(serverSocket)
	return -1
    end if

    -- connect to the server    
    if -1 = c_func(connect, {sd, serverSocket, sizeof_sockaddr}) then
    	c_proc(perror, {alloc_string("connect")})
	free_strings(1)
	free(serverSocket)
    	return -1
    end if
    
    return sd
end function

global procedure remote_close(integer sd)
    c_proc(close_, {sd})
    free(serverSocket)
end procedure


global function remote_routine_id(integer sd, sequence name)
    write_object(sd, name)
    return read_object(sd)
end function

global procedure remote_call_proc(integer sd, integer id, sequence args)
    write_object(sd, id+1)
    write_object(sd, args)
end procedure

global function remote_call_func(atom sd, atom id, sequence args)
    write_object(sd, -id)
    write_object(sd, args)
    return read_object(sd)
end function



